# This package contains routines to validate an XML document using 
# the XML spec. 

# It is based on the spec of 08-Dec-97

package XML::Validate;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
use XML::Char;

require Exporter;
require AutoLoader;

@ISA = qw(Exporter AutoLoader);
@EXPORT = qw( );
@EXPORT_OK = qw ( validate );
$VERSION = '0.01';


# Preloaded methods go here.











# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
=head1 NAME

XML::Validate - Validate an XML document

=head1 SYNOPSIS

  use XML::Validate;
  $p = XML::Validate->new;  # should really a be subclass
  $p->parse($chunk1);
  $p->parse($chunk2);
  #...
  $p->eof;                 # signal end of document

  # Parse directly from file
  $p->parse_file("foo.xml");
  # or
  open(F, "foo.xml") || die;
  $p->parse_file(\*F);

=head1 DESCRIPTION

XML::Validate checks an XML document against both the XML spec and the DTD.

$p->parse() - will tokenize an aribtrary chunk of XML
              document. Returns a reference to the parser object.

$p->eof()   - ends the document and flushes remaining text

$p->parse_file() - parses text from a file. May be passed a filename
                   or an already opened filehandle. Returns a
                   reference to the parser object.

In order to make the parser do anything interesting, you must make a
subclass where you override one or more of the following methods as
appropriate:

=over 4

=item $self->declaration($decl)

This method is called when a I<markup declaration> has been
recognized.  For typical HTML documents, the only declaration you are
likely to find is <!DOCTYPE ...>.  The initial "<!" and ending ">" is
not part of the string passed as argument.  Comments are removed and
entities have B<not> been expanded yet.

=item $self->start($tag, $attr, $attrseq, $origtext)

This method is called when a complete start tag has been recognized.
The first argument is the tag name (in lower case) and the second
argument is a reference to a hash that contain all attributes found
within the start tag.  The attribute keys are converted to lower case.
Entities found in the attribute values are already expanded.  The
third argument is a reference to an array with the lower case
attribute keys in the original order.  The fourth argument is the
original HTML text.


=item $self->end($tag)

This method is called when an end tag has been recognized.  The
argument is the lower case tag name.

=item $self->text($text)

This method is called when plain text in the document is recognized.
The text is passed on unmodified and might contain multiple lines.
Note that for efficiency reasons entities in the text are B<not>
expanded.  You should call HTML::Entities::decode($text) before you
process the text any further.

=item $self->comment($comment)

This method is called as comments are recognized.  The leading and
trailing "--" sequences have been stripped off the comment text.

=back

The default implementation of these methods does nothing, I<i.e.,> the
tokens are just ignored.

There is really nothing in the basic parser that is HTML specific, so
it is likely that the parser can parse many kinds of SGML documents,
but SGML has many obscure features (not implemented by this module)
that prevent us from renaming this module as C<SGML::Parse>.

=head1 SEE ALSO

L<HXML::WellFormed>

=head1 COPYRIGHT

Copyright 1996 Gisle Aas. All rights reserved.

This library is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 AUTHOR

Mark Kvale <kvale@phy.ucsf.edu>

=cut
